function encode_thing(x)
	if type(x) == "number" then return tostring(x)
	elseif type(x) == "string" then return string.format("%q", x):gsub("\\\n", "\n")
	elseif type(x) == "boolean" then
		if x == true then
			return "True"
		else
			return "False"
		end
	elseif type(x) == "table" then
		return table_to_str(x)
	end
end

function decode_string(x) -- TODO-future: can this be done in a better way?
	f = loadstring("return " .. x)
	if f then
		return f()
	else
		return nil
	end
end

function decode_thing(x)
	local num = tonumber(x)
	if num then return num end
	if x == "True" then return true end
	if x == "False" then return false end
--~ 	if x == nil then -- TODO: Hack, it's a bug when this happens
--~ 		print("Error decoding string, x = nil")
--~ 		return "nil"
--~ 	end
	local str = x:match([[^"(.*)"$]]) -- TODO-future: redo these in some other way?
	if str then
		s = decode_string(x)
		if s then return s end
	end
	local tblstr = x:match([[^(%[.+%])$]]) -- TODO-future: redo these in some other way?
	if tblstr then return str_to_table(tblstr) end
	local eff1, eff2, eff3 = x:match([[^%("(.-)", *([0-9]-), *"(.-)"%)$]])
	if eff1 and eff2 and eff3 then
		return { eff1, eff2, eff3 }
	end
	print("error: can't decode table value", x)
end

function table_to_str(tbl)
	local newtbl = {}
	local keys = {}
	for from, to in pairs(tbl) do
		table.insert(keys, from)
	end
	table.sort(keys)
	for _, from in pairs(keys) do
		local to = tbl[from]
--~ 		print("tbl["..tostring(from).."] -> "..tostring(to))
		table.insert(newtbl, string.format("(%s, %s)", encode_thing(from), encode_thing(to)))
	end
	return "["..table.concat(newtbl, ", ").."]"
end

function str_to_table(str)
	local tbl = {}
--~ 	print("str_to_table("..str..")")
	for x in string.gmatch(str, "%b()") do
--~ 		print(x)
		a, b = x:match([[^%(("Go Get 'Em, Tiger!"), *([0-9]+)%)$]]) -- ugly hack for Go Get 'Em, Tiger!
		if not b then
			a, b = x:match([[^%((.-), *(.+)%)$]]) -- TODO-future: redo regex?
		end
--~ 		print("decoding", a, "->", b)
		tbl[decode_thing(a)] = decode_thing(b)
	end
--~ 	print("returning", table_to_str(tbl))
	return tbl
end

function parse_params(str)
	local loadstr = parse_table_string(str)
	if loadstr == "" then
		return {}
	else
		return assert(loadstring(loadstr))()
	end
end

function printstr(tbl)
	if tbl == nil then
		return tostring(tbl)
	elseif type(tbl) == "table" then
		return table_to_str(tbl)
	else
		return tostring(tbl)
	end
end

function get_daily_counter(name)
	local value = get_day_state(name)
	if (value == "") then value = 0 end
	return tonumber(value)
end

function increase_daily_counter(name, amount)
	local increase = (amount or 1)
	set_day_state(name, get_daily_counter(name) + increase)
end

local familiar_counters = {}
local infolines = {}
local values = {}

function print_familiar_counter(compact, normal)
	if string.match(text, "<!%-%- charpane compact") then
		table.insert(familiar_counters, compact)
	else
		table.insert(familiar_counters, normal)
	end
end

function print_charpane_infoline(compact, normal)
	if string.match(text, "<!%-%- charpane compact") then
		table.insert(infolines, compact)
	else
		table.insert(infolines, normal)
	end
end

function print_charpane_value(value)
	table.insert(values, value)
end

function print_charpane_lines()
	if text:match("http://images.kingdomofloathing.com/otherimages/inf_small.gif") then return end
	if string.match(text, "<!%-%- charpane compact") then -- compact mode
		local lines = {}
		for x,y in pairs(familiar_counters) do table.insert(lines, y) end
		for x,y in pairs(infolines) do table.insert(lines, y) end

		for x,y in pairs(lines) do
			if string.match(text, "</table>(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)") then
				text = string.gsub(text, "(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)", function(one) return y .. one end)
			else
				text = string.gsub(text, "(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)", function(one) return "<br>" .. y .. one end)
			end
		end
		for x,y in pairs(values) do
			local name = y["name"] or "{nil}"
			local value = y["value"] or "{nil}"
			local ct = "<td align=right>" .. name .. ":</td><td align=left><b>" .. value .. "</b></td>"
			local style = ""
			if y["color"] then
				style = " style=\"color:" .. y["color"] .. "\""
			end
			if y["link"] then
				ct = [[<td align=right><a target="mainpane" href="]] .. y["link"] .. [["]] .. style .. [[>]] .. name .. [[</a>:</td><td align=left><a target="mainpane" href="]] .. y["link"] .. [["]] .. style .. [[><b>]] .. value .. [[</b></a></td>]]
			end
			if y["tooltip"] then
				ct = "<tr title=\"" .. y["tooltip"] .. "\"" .. style .. ">" .. ct .. "</tr>"
			else
				ct = "<tr" .. style .. ">" .. ct .. "</tr>"
			end
			text = string.gsub(text, "(<!%-%- charpane compact text space %-%->)", function(one) return ct .. one end)
		end
	else -- normal mode
		local lines = {}
		for x,fc in pairs(familiar_counters) do table.insert(lines, tostring(fc)) end
		for x,il in pairs(infolines) do table.insert(lines, tostring(il)) end

		for x,l in pairs(lines) do
			text = string.gsub(text, "(<!%-%- charpane normal familiar text space type%b{} weight%b{} %-%->)", function(one) return "<tr><td colspan=\"2\"><center><font size=\"2\">" .. l .. "</font></center></td></tr>" .. one end)
		end
		for x,y in pairs(values) do
			local name = y["name"] or "{nil}"
			local value = y["value"] or "{nil}"
			local nt = "<font size=\"2\">" .. name .. ": <b>" .. value .. "</b></font>"
			local style = ""
			if y["color"] then
				style = " style=\"color:" .. y["color"] .. "\""
			end
			if y["link"] then
				nt = [[<font size="2"><a target="mainpane" href="]] .. y["link"] .. [["]] .. style .. [[>]] .. name .. [[: <b>]] .. value .. [[</b></a></font>]]
			end
			if y["tooltip"] then
				nt = "<span title=\"" .. y["tooltip"] .. "\"" .. style .. ">" .. nt .. "</span>"
			else
				nt = "<span" .. style .. ">" .. nt .. "</span>"
			end
			text = string.gsub(text, "(<!%-%- charpane normal text space %-%->)", function(one) return nt .. "<br>" .. one end)
		end
	end
end

function get_mainstat()
	classname = get_session_state("class")
	if classname == "1" or classname == "2" then
		return "Muscle"
	elseif classname == "3" or classname == "4" then
		return "Mysticality"
	elseif classname == "5" or classname == "6" then
		return "Moxie"
	end
end

--~ function get_request_state(name) return get_state("request", name) end
function get_fight_state(name) return get_state("fight", name) end
function get_session_state(name) return get_state("session", name) end
function get_day_state(name) return get_state("day", name) end
function get_ascension_state(name, value) return get_state("ascension", name) end
function get_character_state(name) return get_state("character", name) end

--~ function set_request_state(name, value) set_state("request", name, value) end
function set_fight_state(name, value) set_state("fight", name, value) end
function set_session_state(name, value) set_state("session", name, value) end
function set_day_state(name, value) set_state("day", name, value) end
function set_ascension_state(name, value) set_state("ascension", name, value) end

function get_page(url, params) return submit_page("GET", url, params) end

function post_page(url, params) return submit_page("POST", url, params) end

function get_itemid(name)
	return tonumber(str_to_table(get_item_data_by_name(name)).id)
end

function caching_get_itemid(name) -- TODO: would be better it get_item_data_by_name wasn't slow
	local tbl = get_session_state("cached itemids")
	if (tbl == "") then tbl = {} else tbl = str_to_table(tbl) end
	if not tbl[name] then
		tbl[name] = tonumber(str_to_table(get_item_data_by_name(name)).id)
		set_session_state("cached itemids", table_to_str(tbl))
	end
	return tbl[name]
end

if set_state then
	get_itemid = caching_get_itemid
end







dod_potion_types = {
	"bubbly potion",
	"cloudy potion",
	"dark potion",
	"effervescent potion",
	"fizzy potion",
	"milky potion",
	"murky potion",
	"smoky potion",
	"swirly potion",
}

dod_potion_effects = {
	"acuity",
	"blessing",
	"booze",
	"confusion",
	"detection",
	"healing",
	"sleep",
	"strength",
	"teleportation",
}

function get_dod_potion_status()
--~ 	print("get potion status")
	local tbl = get_ascension_state("zone.dod.potions")
	if (tbl == "") then tbl = {} else tbl = str_to_table(tbl) end
	unknown_potions = {}
	for _, pot in pairs(dod_potion_types) do
		if not tbl[pot] then
			table.insert(unknown_potions, pot)
		end
	end

	found = {}
	for pot, eff in pairs(tbl) do
		found[eff] = pot
	end

	unknown_effects = {}
	for _, eff in pairs(dod_potion_effects) do
		if not found[eff] then
			table.insert(unknown_effects, eff)
		end
	end

	if table.maxn(unknown_effects) == 1 and table.maxn(unknown_potions) == 1 then -- only one missing effect and one missing potion
		tbl[unknown_potions[1]] = unknown_effects[1]
		return tbl, {}
	else
--~ 		print("unknown effects", table.maxn(unknown_effects), table_to_str(unknown_effects))
--~ 		print("unknown potions", table.maxn(unknown_potions), table_to_str(unknown_potions))
		return tbl, unknown_potions
	end
end




function get_bufflist()
	local status = str_to_table(get_status_info())
	local effects = str_to_table(status.effects)
	local buffslist = {}
	for _, x in pairs(effects) do
		buffslist[x[1]] = tonumber(x[2])
	end
	return buffslist
end


function get_drunkenness()
--~ 	local drunk = tonumber(tbl["drunkenness"])
	return 0 -- TODO: hack!
--~ 	return drunk
end

function do_we_have_mall_access()
	return true -- TODO: hack!
end
